"""Command line interface."""

import sys
from pathlib import Path

import click

from .sqlpygen import generate


@click.command()
@click.option(
    "-i",
    "--input",
    "input_file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
    help="Annotated sql file.",
)
@click.option(
    "-o",
    "--output",
    "output_file",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    required=True,
    help="Generated python file.",
)
@click.option(
    "-v", "--verbose",
    is_flag=True,
    help="Print out intermediate results.")
def cli(input_file, output_file, verbose):
    """SqlPyGen

    Generated type annotated python code from annotated SQL
    """
    input_file = Path(input_file)
    output_file = Path(output_file)

    input_ = input_file.read_text()
    try:
        output = generate(input_, verbose)
    except RuntimeError as e:
        click.secho(str(e), fg="red")
        sys.exit(1)

    output_file.write_text(output)
    click.secho("Python Code generated successfully", fg="green")



"""Generate type annotated python from SQL."""

from pprint import pprint
from importlib.resources import read_text

import black
from lark import Lark, Transformer, UnexpectedToken
from jinja2 import Template, StrictUndefined, TemplateSyntaxError


class SqlPyGenTransformer(Transformer):
    """Transform the parse tree for code generation."""

    CNAME = str
    MODULE_NAME = str

    def SQL_STRING(self, t):
        return t.strip().rstrip(";").strip()

    def type(self, ts):
        name, old_type = ts
        return ("types", {"name": name, "old_type": old_type})

    def module(self, ts):
        (name,) = ts
        return ("module", name)

    def pname_ptype(self, ts):
        pname, ptype = ts
        return {"name": pname, "type": ptype}

    def params(self, ts):
        return ("params", ts)

    def return_(self, ts):
        return ("return_", ts)

    def schema(self, ts):
        name, sql = ts
        return ("schemas", {"name": name, "sql": sql})

    def query(self, ts):
        name, sql = ts[0], ts[-1]
        params, return_ = [], []
        for typ, val in ts[1:-1]:
            if typ == "params":
                params = val
            elif typ == "return_":
                return_ = val
            else:
                raise ValueError(f"Unexpected child type: {typ=} {val=}")

        return (
            "queries",
            {"name": name, "params": params, "return_": return_, "sql": sql},
        )

    def import_stmt(self, ts):
        module, *names = ts
        return ("import_stmts", {"module": module, "names": names})

    def start(self, ts):
        ret = {
            "module": None,
            "import_stmts": [],
            "types": [],
            "schemas": [],
            "queries": [],
        }
        for grp, val in ts:
            if grp == "module":
                ret["module"] = val
            else:
                ret[grp].append(val)

        return ret


def get_parser() -> Lark:
    """Return the parser."""
    grammar = read_text("sqlpygen", "sqlpygen.lark")
    parser = Lark(grammar, parser="lalr")
    return parser


def get_template() -> Template:
    """Return the code generation template."""
    tpl_text = read_text("sqlpygen", "sqlpygen.jinja2")
    try:
        tpl_obj = Template(
            tpl_text, trim_blocks=True, lstrip_blocks=True, undefined=StrictUndefined
        )
    except TemplateSyntaxError as e:
        raise ValueError(f"Syntax error in template line {e.lineno}") from e
    return tpl_obj


def generate(text: str, verbose: bool = False) -> str:
    """Generate python from annotated sql."""
    parser = get_parser()
    transformer = SqlPyGenTransformer()
    template = get_template()

    try:
        parse_tree = parser.parse(text)
    except UnexpectedToken as e:
        line, col = e.line - 1, e.column - 1
        col_m1 = max(0, col)
        err_line = text.split("\n")[line]
        err_marker = "-" * col_m1 + "^"
        msg = f"Error parsing input:\n{e}\n{err_line}\n{err_marker}"
        raise RuntimeError(msg)

    if verbose:
        print("Parse tree")
        print("-" * 80)
        print(parse_tree.pretty())

    trans_tree = transformer.transform(parse_tree)

    if verbose:
        print("Transformed tree")
        print("-" * 80)
        pprint(trans_tree)

    rendered_tree = template.render(**trans_tree)
    rendered_tree = black.format_str(rendered_tree, mode=black.Mode())
    return rendered_tree



"""example1

This module has been generated with SqlPyGen.
"""

from pprint import pprint
from typing import NewType, Iterable, cast

import sqlite3

ConnectionType = sqlite3.Connection


SCHEMA = {}
SCHEMA[
    "table_stocks"
] = """
CREATE TABLE stocks (
    date text,
    trans text,
    symbol text,
    qty real,
    price real
)
"""


QUERY = {}
QUERY[
    "insert_into_stocks"
] = """
INSERT INTO stocks VALUES (:date, :trans, :symbol, :qty, :price)
"""

QUERY[
    "select_from_stocks"
] = """
SELECT * FROM stocks
"""


def create_schema(connection: ConnectionType) -> None:
    """Create the table schema."""
    with connection:
        cursor = connection.cursor()

        try:
            sql = SCHEMA["table_stocks"]

            cursor.execute(sql)
        except Exception as e:
            raise RuntimeError(
                "An unexpected exception occurred when creating schema: table_stocks"
            ) from e


def insert_into_stocks(
    connection: ConnectionType,
    date: str,
    trans: str,
    symbol: str,
    qty: float,
    price: float,
) -> None:
    """Query insert_into_stocks with transaction."""
    query_args = {
        "date": date,
        "trans": trans,
        "symbol": symbol,
        "qty": qty,
        "price": price,
    }

    with connection:
        cursor = connection.cursor()
        try:
            sql = QUERY["insert_into_stocks"]

            cursor.execute(sql, query_args)

        except Exception as e:
            raise RuntimeError(
                "An unexpected exception occurred while executing query: insert_into_stocks"
            ) from e


def insert_into_stocks_nt(
    connection: ConnectionType,
    date: str,
    trans: str,
    symbol: str,
    qty: float,
    price: float,
) -> None:
    """Query insert_into_stocks no implied transaction."""
    query_args = {
        "date": date,
        "trans": trans,
        "symbol": symbol,
        "qty": qty,
        "price": price,
    }

    cursor = connection.cursor()
    try:
        sql = QUERY["insert_into_stocks"]

        cursor.execute(sql, query_args)

    except Exception as e:
        raise RuntimeError(
            "An unexpected exception occurred while executing query: insert_into_stocks"
        ) from e


def select_from_stocks(
    connection: ConnectionType,
) -> list[tuple[str, str, str, float, float]]:
    """Query select_from_stocks with transaction."""

    with connection:
        cursor = connection.cursor()
        try:
            sql = QUERY["select_from_stocks"]

            cursor.execute(sql)

            return cursor.fetchall()
        except Exception as e:
            raise RuntimeError(
                "An unexpected exception occurred while executing query: select_from_stocks"
            ) from e


def select_from_stocks_nt(
    connection: ConnectionType,
) -> Iterable[tuple[str, str, str, float, float]]:
    """Query select_from_stocks no implied transaction."""

    cursor = connection.cursor()
    try:
        sql = QUERY["select_from_stocks"]

        cursor.execute(sql)

        return cast(Iterable[tuple[str, str, str, float, float]], cursor)
    except Exception as e:
        raise RuntimeError(
            "An unexpected exception occurred while executing query: select_from_stocks"
        ) from e


def explain_queries() -> None:
    connection = sqlite3.connect(":memory:")
    create_schema(connection)

    with connection:
        cursor = connection.cursor()

        try:
            sql = QUERY["insert_into_stocks"]
            sql = "EXPLAIN " + sql

            query_args = {
                "date": None,
                "trans": None,
                "symbol": None,
                "qty": None,
                "price": None,
            }

            cursor.execute(sql, query_args)

            print("Query explanation for insert_into_stocks")
            print("-" * 80)
            pprint(cursor.fetchall())
        except Exception as e:
            raise RuntimeError(
                "An unexpected exception occurred while executing query plan for: insert_into_stocks"
            ) from e

        try:
            sql = QUERY["select_from_stocks"]
            sql = "EXPLAIN " + sql

            cursor.execute(sql)

            print("Query explanation for select_from_stocks")
            print("-" * 80)
            pprint(cursor.fetchall())
        except Exception as e:
            raise RuntimeError(
                "An unexpected exception occurred while executing query plan for: select_from_stocks"
            ) from e


if __name__ == "__main__":
    explain_queries()
